@@ -28,7 +28,7 @@ function run_test {
 	else
 		echo "== Test $1"
 		mkdir wdir/i/$1/
-		if ! ./$1 wdir/s/cs.sk wdir/i/$1/; then
+		if ! setsid ./$1 wdir/s/cs.sk wdir/i/$1/ < /dev/null &>> wdir/i/$1/test.log; then
 			echo "$1: FAIL"
 			RESULT=1
 		fi
@@ -39,8 +39,10 @@ run_test test_sub
 run_test test_self
 run_test test_notify
 run_test test_iters
+run_test test_errno
 
 echo "== Stopping service"
 kill -TERM $(cat wdir/s/pidfile)
+unlink libcriu.so.1
 [ $RESULT -eq 0 ] && echo "Success" || echo "FAIL"
 exit $RESULT
